/*
 * Copyright 2012-2018 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.tgau;

import java.io.PrintStream;
import java.io.StringReader;
import java.util.Arrays;

import at.jku.risc.stout.tgau.algo.AntiUnify;
import at.jku.risc.stout.tgau.algo.AntiUnifySystem;
import at.jku.risc.stout.tgau.algo.DebugLevel;
import at.jku.risc.stout.tgau.algo.RigidityFnc;
import at.jku.risc.stout.tgau.algo.RigidityFncSubsequence;
import at.jku.risc.stout.tgau.data.InputParser;
import at.jku.risc.stout.tgau.data.NodeFactory;
import at.jku.risc.stout.tgau.data.TermGraph;

public class AntiUnifyMain {

	static DebugLevel debugLevel = DebugLevel.SIMPLE;
	static boolean iterateAll = false;
	static boolean justifySigma = false;
	static boolean normalize = false;
	static int minLen = 1;
	static String rigidityFnc = "at.jku.risc.stout.tgau.algo.RigidityFncSubsequence";
	static String commutativeSymbols = null;
	private static boolean splitRule = false;

	public static void main(String[] args) {
		if (args.length >= 2) {
			if ("-d".equals(args[0])) {
				debugLevel = DebugLevel.PROGRESS;
				main(Arrays.copyOfRange(args, 1, args.length));
			} else if ("-v".equals(args[0])) {
				if (debugLevel != DebugLevel.PROGRESS)
					debugLevel = DebugLevel.VERBOSE;
				main(Arrays.copyOfRange(args, 1, args.length));
			} else if ("-r".equals(args[0])) {
				if (debugLevel == DebugLevel.SIMPLE)
					debugLevel = DebugLevel.RESULT;
				main(Arrays.copyOfRange(args, 1, args.length));
			} else if ("-a".equals(args[0])) {
				iterateAll = true;
				main(Arrays.copyOfRange(args, 1, args.length));
			} else if ("-n".equals(args[0])) {
				normalize = true;
				main(Arrays.copyOfRange(args, 1, args.length));
			} else if ("-s".equals(args[0])) {
				if (commutativeSymbols != null) {
					System.out.println("The split rule is not supported in combination with commutative functions.");
				} else {
					splitRule = true;
					main(Arrays.copyOfRange(args, 1, args.length));
				}
			} else if ("-h".equals(args[0])) {
				printHelp(System.out);
			} else if ("-j".equals(args[0])) {
				justifySigma = true;
				System.out.println("This argument has not been implemented yet!");
				System.out.println();
				printHelp(System.out);
			} else if ("-c".equals(args[0])) {
				if (splitRule) {
					System.out.println("The split rule is not supported in combination with commutative functions.");
				} else if (args.length < 2) {
					System.out.println("Insufficient arguments!");
					System.out.println();
					printHelp(System.out);
				} else {
					commutativeSymbols = args[1];
					main(Arrays.copyOfRange(args, 2, args.length));
				}
			} else if ("-f".equals(args[0])) {
				if (args.length < 2) {
					System.out.println("Insufficient arguments!");
					System.out.println();
					printHelp(System.out);
				} else {
					rigidityFnc = args[1];
					main(Arrays.copyOfRange(args, 2, args.length));
				}
			} else if ("-m".equals(args[0])) {
				if (args.length < 2) {
					System.out.println("Insufficient arguments!");
					System.out.println();
					printHelp(System.out);
				} else {
					minLen = Integer.parseInt(args[1]);
					main(Arrays.copyOfRange(args, 2, args.length));
				}
			} else {
				try {
					RigidityFnc r = (RigidityFnc) Class.forName(rigidityFnc).newInstance();
					r.setMinLen(minLen);
					RigidityFncSubsequence rC;
					if (r instanceof RigidityFncSubsequence) {
						rC = (RigidityFncSubsequence) r;
					} else {
						rC = new RigidityFncSubsequence();
						rC.setMinLen(minLen);
					}
					NodeFactory nf = new NodeFactory();
					boolean comm = commutativeSymbols != null;
					if (comm) {
						nf.setCommutative(commutativeSymbols.trim().split("\\s+"));
						AntiUnifySystem.ENABLE_COMMUTATIVE_RULES = true;
					}
					AntiUnifySystem.ENABLE_SPLIT_RULE = splitRule;
					TermGraph g1 = new InputParser(nf).parseTermGraph(new StringReader(args[0]), comm);
					TermGraph g2 = new InputParser(nf).parseTermGraph(new StringReader(args[1]), comm);
					AntiUnify rau = new AntiUnify(r, rC, g1, g2, nf);
					rau.antiUnify(iterateAll, normalize, debugLevel, System.out);

					if (debugLevel != DebugLevel.RESULT) {
						System.out.println();
						System.out.println("Supported by the Austrian Science Fund (FWF) under the project J 3909-N31");
					}
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		} else {
			System.out.println("Insufficient arguments!");
			System.out.println();
			printHelp(System.out);
		}
	}

	private static void printHelp(PrintStream out) {
		System.out.println("Anti-unification problem solver for term graphs");
		System.out.println();
		System.out.println("Give two term graphs of the form:");
		System.out.println("\"{x0 = t0, ..., xn = tn, X0 = Y0, ..., Xm = Ym}\"");
		System.out.println(
				"where x0,...,xn are term recursion variables, t0,...,tn are terms, X0,...,Xm are hedge recursion variables, and Y0,...,Ym hedge variables");
		System.out.println(
				"Input example: \"{x0 = f(x1,x2,x3,x0,x3,x2, x3); x1 = g(x1,x2); x2 = b; x3 = a}\" \"{y0 = f(y1, y0, y3),y1 = g(y1, y2), y2= b, y3 = a}\"");
		System.out.println();
		System.out.println("Simple execution example: " + "java at.jku.risc.stout.urau.AntiUnifyMain -a "
				+ "\"{x0 = f(x1,x2,x3,x0,x3,x2, x3); x1 = g(x1,x2); x2 = b; x3 = a}\" \"{y0 = f(y1, y0, y3),y1 = g(y1, y2), y2= b, y3 = a}\"");
		System.out.println("Advanced execution example with jvm parameters:");
		System.out.println(
				"java -Xmx2048m -XX:MaxPermSize=1024m -Xss128m " + "at.jku.risc.stout.urau.AntiUnifyMain -a -m 3 -f "
						+ "\"at.jku.risc.stout.tgau.algo.RigidityFncSubstring\" G1 G2");
		System.out.println("G1 and G2 denote some graphs of the above described form.");
		System.out.println();
		System.out.println("  -a    iterate all possibilities");
		// System.out.println(" -j justify the computed generalization");
		System.out.println("  -f x  set rigidity function to x where x is the full class name (with package name)");
		System.out.println("  -m x  set minimum alignment length to x");
		System.out.println(
				"  -n    normalize the result. I.e., rename all the variables uniformly and remove duplicates");
		System.out.println("  -v    verbose output mode prints sigma and the store");
		System.out.println("  -r    print only the result without any text (minimum output).");
		System.out.println("  -d    debug mode prints out every available information");
		System.out.println(
				"  -c x  enable commutative mode and make the symbols that appear in x commutative. E.g.: -c \"f1 f2 f3\"");
		System.out.println("  -s    enable the split rule that converts hedge variables to term variables if possible");
		System.out.println("  -h    print this information");
		System.out.println();
		System.out.println("Supported by the Austrian Science Fund (FWF) under the project J 3909-N31");
	}
}
