/*
 * Copyright 2012-2018 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.tgau.algo;

import java.io.PrintStream;
import java.util.Deque;
import java.util.Queue;
import java.util.Set;

import at.jku.risc.stout.tgau.data.NodeFactory;
import at.jku.risc.stout.tgau.data.TermGraph;
import at.jku.risc.stout.tgau.data.atom.TermVar;
import at.jku.risc.stout.tgau.util.DataStructureFactory;
import at.jku.risc.stout.tgau.util.PrintableX;

/**
 * This class encapsulates the rule based system {@linkplain AntiUnifySystem}
 * and takes care of the system branching.
 * 
 * @author Alexander Baumgartner
 */
public class AntiUnify {
	private Queue<AntiUnifySystem> eqBranch = DataStructureFactory.$.newDeque();
	private Set<TermGraph> normalizedSet;

	/**
	 * Attention: Do not run two instances in parallel, since
	 * {@linkplain AntiUnifySystem} uses global static arguments!
	 */
	public AntiUnify(RigidityFnc rigidFnc, RigidityFncSubsequence rigidFncCommutative, TermGraph g1, TermGraph g2,
			NodeFactory nf) {
		AntiUnifySystem.BRANCH_COUNT = 0; // reset branch counter
		AntiUnifySystem.G1 = g1; // make g1 globally available
		AntiUnifySystem.G2 = g2; // make g2 globally available
		AntiUnifySystem.rFnc = rigidFnc;
		AntiUnifySystem.rFncC = rigidFncCommutative;
		AntiUnifySystem.nf = nf;
		AntiUnifySystem.branchPointer = this.eqBranch;

		TermVar startVar = nf.obtainFreshTermVar();
		TermGraph g = new TermGraph(startVar);
		// TODO: Paper: Construct g with a most general recursion equation.
		Deque<PrintableX> eq = DataStructureFactory.$.newDeque();
		eq.add(new TermAUP(startVar, g1.getRoot(), g2.getRoot()));
		this.eqBranch.add(new AntiUnifySystem(eq, DataStructureFactory.$.<HedgeAUP>newList(),
				DataStructureFactory.$.<HedgeAUP>newList(), DataStructureFactory.$.<TermAUP>newList(),
				DataStructureFactory.$.<TermAUP, TermVar>newMap(), g));
	}

	/**
	 * Silently computes the generalizations. If the given argument is false then
	 * only one result will be computed, otherwise all generalizations are computed.
	 */
	public long antiUnify(boolean iterateAll, boolean normalize) throws IllegalAlignmentException {
		return antiUnify(iterateAll, normalize, DebugLevel.SILENT, null);
	}

	/**
	 * Start computation with the specified {@linkplain PrintStream} to show the
	 * progress of the computation. If the first argument is false then only one
	 * result will be computed, otherwise all generalizations are computed.
	 */
	public long antiUnify(boolean iterateAll, boolean normalize, DebugLevel debugLevel, PrintStream out)
			throws IllegalAlignmentException {
		AntiUnifySystem.debugLevel = debugLevel;
		AntiUnifySystem.debugOut = out;
		TermGraphNormalizer tgn = null;
		if (normalize) {
			tgn = new TermGraphNormalizer();
			normalizedSet = DataStructureFactory.$.newSet();
		}
		while (!eqBranch.isEmpty()) {
			AntiUnifySystem sys = eqBranch.poll();
			TermGraph result = sys.compute();
			if (!normalize || debugLevel != DebugLevel.RESULT)
				printDebug(debugLevel, out, sys);
			if (normalize) {
				result = tgn.normalize(result);
				if (normalizedSet.add(result) && debugLevel == DebugLevel.RESULT)
					out.println(result);
			}
			callback(result);
			if (!iterateAll)
				break;
		}
		if (debugLevel != DebugLevel.SILENT && debugLevel != DebugLevel.RESULT) {
			if (!iterateAll && AntiUnifySystem.BRANCH_COUNT > 1)
				out.println("More generalizations available");
			else
				out.println(AntiUnifySystem.BRANCH_COUNT + " generalizations found");
			if (normalize) {
				out.println();
				out.println(normalizedSet.size() + " normalized results:");
				for (TermGraph result : normalizedSet) {
					out.println(result);
				}
			}
		}
		return AntiUnifySystem.BRANCH_COUNT;
	}

	public void printDebug(DebugLevel debugLevel, PrintStream out, AntiUnifySystem sys) {
		switch (debugLevel) {
		case VERBOSE:
			out.println();
			out.println("  Store " + sys.getBranchId() + ": " + sys.getStore());
			out.print("  Trail " + sys.getBranchId() + ": ");
			sys.printTrail();
		case PROGRESS:
			out.println("-----------");
		case SIMPLE:
			out.println(" Result " + sys.getBranchId() + ": " + sys.getGraph());
			out.println("-----------");
			break;
		case RESULT:
			out.println(sys.getGraph());
		default:
		}
	}

	/**
	 * This callback function will be invoked for every found generalization. By
	 * default it does nothing. For instance, one can use this callback function to
	 * collect all the results or to filter out some interesting results.<br>
	 * In combination with a matching algorithm, one can obtain the minimal complete
	 * set of generalizations by using this callback function.
	 */
	public void callback(TermGraph generalization) {
	}

	/**
	 * The set of all the computed and normalized generalizations.
	 */
	public Set<TermGraph> getNormalizedSet() {
		return normalizedSet;
	}
}
