/*
 * Copyright 2012-2018 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.tgau.algo;

import java.io.PrintStream;
import java.util.Collections;
import java.util.Deque;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Queue;

import at.jku.risc.stout.tgau.algo.AlignmentList.Alignment;
import at.jku.risc.stout.tgau.algo.AlignmentList.Alignment.AlignmentAtom;
import at.jku.risc.stout.tgau.data.NodeFactory;
import at.jku.risc.stout.tgau.data.TermGraph;
import at.jku.risc.stout.tgau.data.atom.FunctionApplication;
import at.jku.risc.stout.tgau.data.atom.HedgeVar;
import at.jku.risc.stout.tgau.data.atom.Term;
import at.jku.risc.stout.tgau.data.atom.TermVar;
import at.jku.risc.stout.tgau.data.atom.Variable;
import at.jku.risc.stout.tgau.util.DataStructureFactory;
import at.jku.risc.stout.tgau.util.IntList;
import at.jku.risc.stout.tgau.util.PrintableX;

/**
 * This class represents a rule based system for rigid anti-unification of term
 * graphs. Parallel executation of two instances is NOT possible, since there
 * are global arguments that are used among all the branches of one and the same
 * run.<br>
 * The algorithm Gen(R) is described in the paper:<br>
 * <a href="http://www.risc.jku.at/home/abaumgar/"> Alexander Baumgartner</a>,
 * <a href="http://www.risc.jku.at/people/tkutsia/"> Temur Kutsia</a>,
 * <a href="http://www.iiia.csic.es/~levy/"> Jordi Levy</a>,
 * <a href="http://ima.udg.edu/~villaret/"> Mateu Villaret</a>. Term-Graph
 * Anti-Unification.
 * 
 * @author Alexander Baumgartner
 */
public class AntiUnifySystem {
	private int branchId = ++BRANCH_COUNT;
	private Deque<PrintableX> problemSet;
	private List<HedgeAUP> storeH;
	private List<HedgeAUP> storeC;
	private List<TermAUP> storeT;
	private List<? super PrintableX> store;
	private Map<TermAUP, TermVar> trail;
	private TermGraph g;

	public static boolean ENABLE_SPLIT_RULE;
	public static boolean ENABLE_COMMUTATIVE_RULES;
	public static NodeFactory nf;
	public static Queue<AntiUnifySystem> branchPointer;
	public static RigidityFnc rFnc;
	public static RigidityFnc rFncC;
	public static TermGraph G1;
	public static TermGraph G2;
	public static int BRANCH_COUNT = 0;
	public static String OUTPUT_SEPARATOR = "; ";
	public static DebugLevel debugLevel;
	public static PrintStream debugOut;

	/**
	 * Most likely you don't need this constructor because the algorithm is
	 * encapsulated in the class {@linkplain AntiUnify} which is much easier to use.
	 * 
	 * @param problemSet
	 *            The set of AUPs to be generalized.
	 * @param storeH
	 *            The none-commutative store keeps track of solved hedge equations
	 *            that are not commutative.
	 * @param storeC
	 *            The commutative store that keeps track of solved hedge equations
	 *            that are commutative.
	 * @param storeT
	 *            The term store that keeps track of solved term equations.
	 * @param trail
	 *            The trail keeps track of recursion variables.
	 * @param g
	 *            The computed generalization. It is a term graph.
	 */
	public AntiUnifySystem(Deque<PrintableX> problemSet, List<HedgeAUP> storeH, List<HedgeAUP> storeC,
			List<TermAUP> storeT, Map<TermAUP, TermVar> trail, TermGraph g) {
		this.problemSet = problemSet;
		this.storeH = storeH;
		this.storeT = storeT;
		this.storeC = storeC;
		this.store = DataStructureFactory.$.<PrintableX>newConcatList(storeT,
				DataStructureFactory.$.<PrintableX>newConcatList(storeH, storeC));
		this.trail = trail;
		this.g = g;
	}

	/**
	 * Most likely you don't need this method because the algorithm is encapsulated
	 * in the class {@linkplain AntiUnify} which is much easier to use.
	 */
	public TermGraph compute() throws IllegalAlignmentException {
		debug(null);
		while (!problemSet.isEmpty()) {
			PrintableX aup = problemSet.removeLast();
			boolean appl = false;
			if (aup instanceof TermAUP) {
				TermAUP tAup = (TermAUP) aup;
				// Check Share
				TermVar termVar = trail.get(tAup);
				if (termVar != null) {
					appl = true;
					g.substitute(tAup.generalizationVar, termVar);
					debug("Share");
				} else {
					Term n1 = G1.get(tAup.left);
					Term n2 = G2.get(tAup.right);
					// Check Step
					// For performance reasons we assume that rFnc returns the two equal symbols and
					// do not call it explicitly here.
					if (n1 instanceof FunctionApplication && n1 != null && n2 != null && n1.getAtom() == n2.getAtom()) {
						appl = true;
						FunctionApplication f1 = (FunctionApplication) n1;
						FunctionApplication f2 = (FunctionApplication) n2;
						boolean commutative = ENABLE_COMMUTATIVE_RULES && f1.getSymbol().isCommutative();
						if (!f1.getArgs().isEmpty() || !f2.getArgs().isEmpty()) {
							HedgeVar hedgeVar = nf.obtainFreshHedgeVar();
							problemSet.add(new HedgeAUP(hedgeVar, f1.getArgs(), f2.getArgs(), commutative));
							List<Variable> l = DataStructureFactory.$.newList();
							l.add(hedgeVar);
							n1 = new FunctionApplication(f1.getSymbol(), l);
						}
						termVar = nf.obtainFreshTermVar();
						trail.put(tAup, termVar);
						g.substitute(tAup.generalizationVar, termVar);
						g.getRecursionEquations().put(termVar, n1);
						if (commutative)
							debug("Step-C");
						else
							debug("Step");
					}
				}
			} else {
				HedgeAUP hAup = (HedgeAUP) aup;
				// Decompose
				AlignmentList align;
				if (hAup.commutative) {
					align = rFncC.compute(G1.top(hAup.left), G2.top(hAup.right));
					for (int i = align.size() - 1; i >= 0; i--) {
						AlignmentList alignC = align.get(i).makeCommutative(hAup.left, hAup.right);
						doDecSBranch(hAup, alignC, 0);
						alignC.free();
					}

				} else {
					align = rFnc.compute(G1.top(hAup.left), G2.top(hAup.right));
				}
				// Check Dec-S
				if (!align.isEmpty()) {
					appl = true;
					doDecSBranch(hAup, align, 1);
					boolean split = this.rigidDecompose(hAup, align.get(0));
					debugDec(hAup.commutative, split, align.get(0), this);
				}
				align.free();
			}
			// Solve if none of the above was applicable
			if (!appl) {
				Variable recV, oldGenV, newGenV;
				String ruleName = "Solve";
				if (aup instanceof TermAUP) {
					recV = nf.obtainFreshTermVar();
					oldGenV = ((TermAUP) aup).generalizationVar;
					storeT.add((TermAUP) aup);
					newGenV = oldGenV;
				} else {
					HedgeAUP hAUP = (HedgeAUP) aup;
					if (ENABLE_SPLIT_RULE && hAUP.left.size() == 1 && hAUP.right.size() == 1
							&& hAUP.left.get(0) instanceof TermVar && hAUP.right.get(0) instanceof TermVar) {
						// PAPER: Convert hedge AUP to term AUP = Split rule
						recV = nf.obtainFreshTermVar();
						oldGenV = hAUP.generalizationVar;
						TermAUP tAup = new TermAUP(nf.obtainFreshTermVar(), (TermVar) hAUP.left.get(0),
								(TermVar) hAUP.right.get(0));
						storeT.add(tAup);
						newGenV = tAup.generalizationVar;
						ruleName = "Solve + Split";
					} else {
						recV = nf.obtainFreshHedgeVar();
						oldGenV = hAUP.generalizationVar;
						addStore(hAUP);
						newGenV = oldGenV;
					}
				}
				g.substitute(oldGenV, recV);
				g.getRecursionEquations().put(recV, newGenV);

				debug(ruleName);
			}
		}
		// Merge
		mergeStore(storeH);
		// Split
		if (ENABLE_SPLIT_RULE) {
			Iterator<HedgeAUP> iter = storeH.iterator();
			// var instead of GEN vars. AND rename t, r to u, v.
			splitall: while (iter.hasNext()) {
				HedgeAUP aup = iter.next();
				if (aup.left.size() == aup.right.size()) {
					for (int i = 0, n = aup.left.size(); i < n; i++) {
						if (aup.left.get(i) instanceof HedgeVar || aup.right.get(i) instanceof HedgeVar)
							continue splitall;
					}
					for (int i = 0, n = aup.left.size(); i < n; i++) {
						TermVar genVar = nf.obtainFreshTermVar();
						TermVar recVar = nf.obtainFreshTermVar();
						TermVar l = (TermVar) aup.left.set(i, recVar);
						TermVar r = (TermVar) aup.right.get(i);
						storeT.add(new TermAUP(genVar, l, r));
						g.suggestMapping(recVar, genVar);
					}
					g.substitute(g.removeByValue(aup.generalizationVar), aup.left);
					g.commitAllMappings();
					iter.remove();
					debug("Split");
				}
			}
		}
		// Merge again
		mergeStore(storeT);
		if (ENABLE_COMMUTATIVE_RULES && !storeC.isEmpty()) {
			for (HedgeAUP aup : storeC) {
				Collections.sort(aup.left);
				Collections.sort(aup.right);
			}
			// Merge-C1 and "small problems" of Merge-C2
			mergeStore(storeC);
			// Merge-C2
			List<HedgeAUP> tmpStore = DataStructureFactory.$.newList(storeH.size());
			for (HedgeAUP aup : storeH) {
				HedgeAUP aupNew = new HedgeAUP(aup.generalizationVar, aup.left, aup.right, false);
				if (aupNew.left.size() > 1) {
					aupNew.left = DataStructureFactory.$.newList(aupNew.left);
					Collections.sort(aupNew.left);
				}
				if (aupNew.right.size() > 1) {
					aupNew.right = DataStructureFactory.$.newList(aupNew.right);
					Collections.sort(aupNew.right);
				}
				tmpStore.add(aupNew);
			}
			IntList mergeIdx = new IntList();
			for (int i = storeC.size() - 1; i >= 0; i--) {
				HedgeAUP aupI = storeC.get(i);
				for (int j = tmpStore.size() - 1; j >= 0; j--) {
					HedgeAUP aupJ = tmpStore.get(j);
					if (aupI.left.equals(aupJ.left) && aupI.right.equals(aupJ.right))
						mergeIdx.add(j);
				}
				while (mergeIdx.size > 0) {
					mergeIdx.size--;
					AntiUnifySystem sysBranch = this;
					if (mergeIdx.size > 0)
						sysBranch = branchSystem();
					sysBranch.storeC.remove(i);
					sysBranch.g.merge(tmpStore.get(mergeIdx.ints[mergeIdx.size]).generalizationVar,
							aupI.generalizationVar);
					sysBranch.debug("Merge-C2");
				}
			}
		}
		return g;
	}

	public void doDecSBranch(HedgeAUP hAup, AlignmentList align, final int idx) throws IllegalAlignmentException {
		for (int i = align.size() - 1; i >= idx; i--) {
			AntiUnifySystem sysBranch = branchSystem();
			boolean split = sysBranch.rigidDecompose(hAup, align.get(i));
			debugDec(hAup.commutative, split, align.get(i), sysBranch);
		}
	}

	private void debugDec(boolean commutative, boolean split, Alignment alignment, AntiUnifySystem sysBranch) {
		String splitS = split ? " + Split  " : "  ";
		if (commutative)
			sysBranch.debug("Dec-S-C" + splitS + alignment);
		else
			sysBranch.debug("Dec-S" + splitS + alignment);
	}

	public void mergeStore(List<? extends BaseAUP> store) {
		for (int i = store.size() - 2; i >= 0; i--) {
			BaseAUP aupI = store.get(i);
			for (int j = store.size() - 1; j > i; j--) {
				BaseAUP aupJ = store.get(j);
				if (aupI.getLeft().equals(aupJ.getLeft()) && aupI.getRight().equals(aupJ.getRight())) {
					store.remove(j);
					g.merge(aupI.getGeneralizationVar(), aupJ.getGeneralizationVar());
					if (aupI.isCommutative() && aupJ.isCommutative()) {
						debug("Merge-C1");
					} else if (aupI.isCommutative() || aupJ.isCommutative()) {
						debug("Merge-C2");
					} else {
						debug("Merge");
					}
				}
			}
		}
	}

	private void debug(String rule) {
		if (debugLevel == DebugLevel.PROGRESS) {
			debugOut.println(rule == null ? "" : rule + " ==> ");
			debugOut.println("Problem " + branchId + ": " + problemSet.toString());
			debugOut.println("  Store " + branchId + ": " + getStore());
			debugOut.print("  Trail " + branchId + ": ");
			printTrail();
			debugOut.println("  Graph " + branchId + ": " + g);
			debugOut.println();
		}
	}

	public void printTrail() {
		debugOut.print('[');
		boolean first = true;
		for (Entry<TermAUP, TermVar> e : trail.entrySet()) {
			if (first) {
				first = false;
			} else {
				debugOut.print(',');
				debugOut.print(' ');
			}
			debugOut.print(e.getValue());
			debugOut.print(':');
			debugOut.print(' ');
			debugOut.print(e.getKey().left);
			debugOut.print(BaseAUP.PRINT_EQ_SEPARATOR);
			debugOut.print(e.getKey().right);
		}
		debugOut.print(']');
		debugOut.println();
	}

	private AntiUnifySystem branchSystem() {
		AntiUnifySystem sysBranch = cloneSystem();
		if (debugLevel == DebugLevel.PROGRESS) {
			debugOut.println("Problem " + branchId + " spawned child problem " + sysBranch.branchId);
			debugOut.println();
		}
		branchPointer.add(sysBranch);
		return sysBranch;
	}

	private AntiUnifySystem cloneSystem() {
		return new AntiUnifySystem(DataStructureFactory.$.newDeque(problemSet), DataStructureFactory.$.newList(storeH),
				DataStructureFactory.$.newList(storeC), DataStructureFactory.$.newList(storeT),
				DataStructureFactory.$.newMap(trail), g.copy());
	}

	/**
	 * The argument aup is not in the current system. Adds the decomposition of aup
	 * to the current system.
	 */
	private boolean rigidDecompose(HedgeAUP aup, Alignment alignment) throws IllegalAlignmentException {
		List<Variable> leftH = aup.left;
		List<Variable> rightH = aup.right;
		boolean appliedSplit = false;

		// Build up substitution hedge while computing the rest
		List<Variable> substiHedge = DataStructureFactory.$.newList();
		// {Y0: v(0-i1)=u(0-j1)} U {Yk: v(ik-ik1)=u(jk-jk1) | 1<=k<=n}
		// U {Yn: v(in-END)=u(jn-END)} U S
		int leftStart = 0, rightStart = 0;
		for (int i = 0, n = alignment.size(); i < n; i++) {
			// This loop includes Y0 and all Yk's
			AlignmentAtom atomK = alignment.get(i);
			if (!aup.commutative) {
				appliedSplit |= composeHedge(substiHedge, leftH, rightH, leftStart, rightStart, atomK.idxLeft,
						atomK.idxRight);
				leftStart = atomK.idxLeft + 1;
				rightStart = atomK.idxRight + 1;
			}
			// decompose the match
			// {Zk: sk=qk | 1<=k<=n} U A
			Variable leftV = leftH.get(atomK.idxLeft);
			Variable rightV = rightH.get(atomK.idxRight);
			if (!(leftV instanceof TermVar) || !(rightV instanceof TermVar))
				throw new IllegalAlignmentException("The term graphs should have disjoint sets of free variables");
			TermVar yk = nf.obtainFreshTermVar();
			problemSet.add(new TermAUP(yk, (TermVar) leftV, (TermVar) rightV));
			substiHedge.add(yk);
		}
		if (!aup.commutative) {
			// The last case is Yn
			appliedSplit |= composeHedge(substiHedge, leftH, rightH, leftStart, rightStart, leftH.size(),
					rightH.size());
		} else {
			// The commutative case
			List<Variable> lH = DataStructureFactory.$.newList(leftH);
			List<Variable> rH = DataStructureFactory.$.newList(rightH);
			for (int k = 0, n = alignment.size(); k < n; k++) {
				AlignmentAtom atomK = alignment.get(k);
				lH.set(atomK.idxLeft, null);
				rH.set(atomK.idxRight, null);
			}
			lH.removeAll(Collections.singleton(null));
			rH.removeAll(Collections.singleton(null));
			if (lH.size() > 0 || rH.size() > 0)
				appliedSplit |= addSubAUP(substiHedge, lH, rH, true);
		}
		// sigma{X -> Y0,f1(Z1),Y1,...,Yn1,fn(Zn),Yn}
		g.substitute(aup.generalizationVar, substiHedge);
		g.commitAllMappings();
		return appliedSplit;
	}

	private boolean composeHedge(List<Variable> substiHedge, List<Variable> leftH, List<Variable> rightH, int leftStart,
			int rightStart, int leftEnd, int rightEnd) {
		if (leftStart == leftEnd && rightStart == rightEnd)
			return false;
		return addSubAUP(substiHedge, leftH.subList(leftStart, leftEnd), rightH.subList(rightStart, rightEnd), false);
	}

	private boolean addSubAUP(List<Variable> substiHedge, List<Variable> lH, List<Variable> rH, boolean commutative) {
		if (ENABLE_SPLIT_RULE && lH.size() == 1 && rH.size() == 1 && lH.get(0) instanceof TermVar
				&& rH.get(0) instanceof TermVar) {
			TermVar Y = nf.obtainFreshTermVar();
			TermVar Z = nf.obtainFreshTermVar();
			substiHedge.add(Z);
			g.suggestMapping(Z, Y);
			storeT.add(new TermAUP(Y, (TermVar) lH.get(0), (TermVar) rH.get(0)));
			return true;
		}
		HedgeVar Y = nf.obtainFreshHedgeVar();
		HedgeVar Z = nf.obtainFreshHedgeVar();
		substiHedge.add(Z);
		g.suggestMapping(Z, Y);
		addStore(new HedgeAUP(Y, lH, rH, commutative));
		return false;
	}

	private void addStore(HedgeAUP aup) {
		if (aup.commutative || (ENABLE_COMMUTATIVE_RULES && aup.left.size() <= 1 && aup.right.size() <= 1))
			storeC.add(aup);
		else
			storeH.add(aup);
	}

	public int getBranchId() {
		return branchId;
	}

	public Deque<PrintableX> getProblemSet() {
		return problemSet;
	}

	/**
	 * A concatenation of {@linkplain #getStoreT()} and {@linkplain #getStoreH()}
	 * and {@linkplain #getStoreC()}
	 */
	public List<? super PrintableX> getStore() {
		return store;
	}

	/**
	 * Store for {@linkplain HedgeAUP}s where the order matters (not commutative
	 * AUPs where one side has at least 2 arguments). If
	 * {@linkplain #ENABLE_COMMUTATIVE_RULES} is false then all
	 * {@linkplain HedgeAUP}s are in this store.
	 */
	public List<HedgeAUP> getStoreH() {
		return storeH;
	}

	/**
	 * Store for {@linkplain HedgeAUP}s where the order does not matter (commutative
	 * AUPs and AUPs that only consist of 0 or 1 arguments). If
	 * {@linkplain #ENABLE_COMMUTATIVE_RULES} is false then this store is empty.
	 */
	public List<HedgeAUP> getStoreC() {
		return storeC;
	}

	/**
	 * Store for {@linkplain TermAUP}s.
	 */
	public List<TermAUP> getStoreT() {
		return storeT;
	}

	public Map<TermAUP, TermVar> getTrail() {
		return trail;
	}

	public TermGraph getGraph() {
		return g;
	}
}
