/*
 * Copyright 2012-2018 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.tgau.algo;

import java.util.List;
import java.util.Map;

import at.jku.risc.stout.tgau.data.TermGraph;
import at.jku.risc.stout.tgau.data.atom.FunctionApplication;
import at.jku.risc.stout.tgau.data.atom.HedgeVar;
import at.jku.risc.stout.tgau.data.atom.Term;
import at.jku.risc.stout.tgau.data.atom.TermVar;
import at.jku.risc.stout.tgau.data.atom.Variable;
import at.jku.risc.stout.tgau.util.DataStructureFactory;

/**
 * This class can be used to normalize the term graphs. That is, all the
 * Variables are renamed uniformly for subsequent calls of the method
 * {@linkplain TermGraphNormalizer#normalize(TermGraph)}.
 * 
 * @author Alexander Baumgartner
 */
public class TermGraphNormalizer {
	public char PREFIX_FreshTermVar = 'x';
	public char PREFIX_FreshHedgeVar = 'Z';

	private List<TermVar> tVarList = DataStructureFactory.$.newList();
	private List<HedgeVar> hVarList = DataStructureFactory.$.newList();
	Map<TermVar, TermVar> tVarMap = DataStructureFactory.$.newMap();
	Map<HedgeVar, HedgeVar> hVarMap = DataStructureFactory.$.newMap();
	private int tVarCount = 0;
	private int hVarCount = 0;

	/**
	 * Normalize all the names of all the variables in the given term graph.
	 */
	public TermGraph normalize(TermGraph tg) {
		if (AntiUnifySystem.ENABLE_COMMUTATIVE_RULES)
			tg.sortCommutative();
		TermGraph tgNew = new TermGraph(normalize(tg.getRoot()));
		normalize(tgNew.getRoot(), tg.getRoot(), tgNew.getRecursionEquations(), tg.getRecursionEquations());
		reset();
		return tgNew;
	}

	private void normalize(Variable recNew, Variable recOld, Map<Variable, Term> mapNew, Map<Variable, Term> mapOld) {
		if (mapNew.containsKey(recNew))
			return;
		Term tOld = mapOld.get(recOld);
		if (tOld instanceof TermVar) {
			mapNew.put(recNew, normalize((TermVar) tOld));
		} else if (tOld instanceof HedgeVar) {
			mapNew.put(recNew, normalize((HedgeVar) tOld));
		} else {
			FunctionApplication applOld = (FunctionApplication) tOld;
			List<Variable> argsNew = DataStructureFactory.$.newList(applOld.getArgs().size());
			mapNew.put(recNew, new FunctionApplication(applOld.getSymbol(), argsNew));
			for (Variable vOld : applOld.getArgs()) {
				if (vOld instanceof TermVar) {
					TermVar vNew = normalize((TermVar) vOld);
					argsNew.add(vNew);
					normalize(vNew, vOld, mapNew, mapOld);
				} else {
					HedgeVar vNew = normalize((HedgeVar) vOld);
					argsNew.add(vNew);
					normalize(vNew, vOld, mapNew, mapOld);
				}
			}
		}
	}

	private TermVar normalize(TermVar x) {
		TermVar renamed = tVarMap.get(x);
		if (renamed == null) {
			if (tVarCount == tVarList.size()) {
				StringBuilder name = new StringBuilder();
				name.append(PREFIX_FreshTermVar);
				name.append(tVarCount);
				renamed = new TermVar(name.toString());
				tVarList.add(renamed);
			} else {
				renamed = tVarList.get(tVarCount);
			}
			tVarCount++;
			tVarMap.put(x, renamed);
		}
		return renamed;
	}

	private HedgeVar normalize(HedgeVar x) {
		HedgeVar renamed = hVarMap.get(x);
		if (renamed == null) {
			if (hVarCount == hVarList.size()) {
				StringBuilder name = new StringBuilder();
				name.append(PREFIX_FreshHedgeVar);
				name.append(hVarCount + 1);
				renamed = new HedgeVar(name.toString());
				hVarList.add(renamed);
			} else {
				renamed = hVarList.get(hVarCount);
			}
			hVarCount++;
			hVarMap.put(x, renamed);
		}
		return renamed;
	}

	public void reset() {
		tVarCount = 0;
		hVarCount = 0;
		tVarMap.clear();
		hVarMap.clear();
	}
}
