/*
 * Copyright 2012-2018 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.tgau.data;

import java.io.IOException;
import java.io.Reader;
import java.util.List;

import at.jku.risc.stout.tgau.data.atom.FunctionSymbol;
import at.jku.risc.stout.tgau.data.atom.HedgeVar;
import at.jku.risc.stout.tgau.data.atom.Term;
import at.jku.risc.stout.tgau.data.atom.TermVar;
import at.jku.risc.stout.tgau.data.atom.Variable;

/**
 * Tiny term parser with the following rules:
 * <ul>
 * <li>[A-Za-z1-9_.] are allowed for variable names and function names. Names
 * can be of any length.
 * <li>If the name is followed by an opening parenthesis a function
 * ({@linkplain FunctionSymbol}) will be generated. Parameter inside the pair of
 * parenthesis are wrapped in a {@linkplain List}
 * <li>Else if the first letter of the name is within [
 * {@linkplain #VARIABLE_START}, z] an {@linkplain TermVar individual variable}
 * will be generated.
 * <ul>
 * <li>If the variable name starts with an upper case letter it will become a
 * {@linkplain HedgeVar hedge variable}. Otherwise an {@linkplain TermVar
 * individual variable} will be instantiates.
 * </ul>
 * <li>Else a function ({@linkplain FunctionSymbol}) with zero parameters
 * (=constant) will be generated.
 * </ul>
 * 
 * @author Alexander Baumgartner
 */
public class InputParser {
	/**
	 * default = 'u' => A word starting with 'u-z' will become a variable.<br>
	 * {@linkplain InputParser#VARIABLE_START} has to be a lower case letter! Hedge
	 * variables are denoted by upper case letters.
	 */
	public static char VARIABLE_START = 'u';
	/**
	 * The parenthesis '(' is used by default for function applications.
	 */
	public static char OPENING_PARENTHESIS = '(';
	/**
	 * The curly (double) bracket '{' is used by default to indicate commutative
	 * function applications.
	 */
	public static char OPENING_PARENTHESIS2 = '{';
	/**
	 * The parenthesis ')' is used by default for function applications.
	 */
	public static char CLOSING_PARENTHESIS = ')';
	/**
	 * The curly (double) bracket '}' is used by default to indicate commutative
	 * function applications.
	 */
	public static char CLOSING_PARENTHESIS2 = '}';
	private int codePoint;
	private NodeFactory nf;

	public InputParser(NodeFactory nf) {
		this.nf = nf;
	}

	/**
	 * Parses a system of recursion equations and returns a term graph. The first
	 * variable is used as the root of the term graph. After parsing/creating the
	 * graph it automatically calls {@linkplain TermGraph#checkAndTrim()} and, if
	 * enableCommutative is true, then also {@linkplain TermGraph#sortCommutative()}
	 * is called. If enableCommutative is true then the parser recognizes
	 * applications that close with {@linkplain #CLOSING_PARENTHESIS2} as
	 * commutative. {@linkplain NodeFactory#setCommutative(String...)} should be
	 * called before parsing the term graphs if some additional symbols should be
	 * treated as commutative.
	 * 
	 * @return The term graph that corresponds to the given set of recursion
	 *         equations.
	 */
	public TermGraph parseTermGraph(Reader in, boolean enableCommutative) throws IOException, MalformedTermException {
		if (!nextNameChar(in))
			return new TermGraph(nf.obtainFreshTermVar());
		Term recVar = parseIt(in, enableCommutative);
		if (!(recVar instanceof TermVar))
			throw new MalformedTermException("Root must be a term variable");
		TermGraph tg = new TermGraph((TermVar) recVar);
		if (!nextNameChar(in))
			return tg;
		tg.addRecursionEquation((Variable) recVar, parseIt(in, enableCommutative));
		while (nextNameChar(in)) {
			recVar = parseIt(in, enableCommutative);
			if (!(recVar instanceof Variable))
				throw new MalformedTermException("Recursion variable expected");
			if (!nextNameChar(in))
				throw new MalformedTermException("No term for recursion variable " + recVar);
			tg.addRecursionEquation((Variable) recVar, parseIt(in, enableCommutative));
		}
		tg.checkAndTrim();
		if (enableCommutative)
			tg.sortCommutative();
		return tg;
	}

	private Term parseIt(Reader in, boolean enableCommutative) throws IOException, MalformedTermException {
		String name = parseTermName(in).toString();
		if (name.length() == 0)
			throw new MalformedTermException("Missing parenthesis at the input term");
		char nameLetter1 = name.charAt(0);
		if (isStartApplication()) {
			nf.pushHedge();
			while (!isEndApplication(in)) {
				if (codePoint == OPENING_PARENTHESIS || codePoint == OPENING_PARENTHESIS2)
					next(in); // consume 2. opening parenthesis
				Term child = parseIt(in, enableCommutative);
				if (child instanceof Variable)
					nf.addToHedge((Variable) child);
				else
					new MalformedTermException("Term graph is not in canonical form");
			}
			if (enableCommutative && codePoint == CLOSING_PARENTHESIS2)
				nf.setCommutative(name); // application of commutative symbol
			List<Variable> hedge = nf.popHedge();
			next(in); // consume closing parenthesis
			if (codePoint == CLOSING_PARENTHESIS || codePoint == CLOSING_PARENTHESIS2)
				next(in); // consume 2. closing parenthesis
			if (Character.toLowerCase(nameLetter1) < VARIABLE_START && nameLetter1 != nf.PREFIX_FreshHedgeVar
					&& nameLetter1 != nf.PREFIX_FreshTermVar)
				return nf.createFunction(name, hedge);
			throw new MalformedTermException("Higher order variables are not supported");
		} else if (Character.toLowerCase(nameLetter1) < VARIABLE_START && nameLetter1 != nf.PREFIX_FreshHedgeVar
				&& nameLetter1 != nf.PREFIX_FreshTermVar) {
			return nf.createConstant(name);
		} else if (nameLetter1 != nf.PREFIX_FreshHedgeVar
				&& (nameLetter1 == nf.PREFIX_FreshTermVar || Character.isLowerCase(nameLetter1))) {
			return nf.createTermVar(name);
		} else {
			return nf.createHedgeVar(name);
		}
	}

	public boolean isEndApplication(Reader in) throws IOException {
		if (codePoint == CLOSING_PARENTHESIS || codePoint == CLOSING_PARENTHESIS2)
			return true;
		next(in); // consume opening parenthesis or comma/semicolon
		return codePoint == CLOSING_PARENTHESIS || codePoint == CLOSING_PARENTHESIS2;
	}

	public boolean isStartApplication() {
		return codePoint == OPENING_PARENTHESIS || codePoint == OPENING_PARENTHESIS2;
	}

	private boolean nextNameChar(Reader in) throws IOException {
		while (!isNameChar(next(in)))
			if (codePoint == -1)
				return false;
		return true;
	}

	private CharSequence parseTermName(Reader in) throws IOException {
		StringBuilder name = new StringBuilder();
		for (; isNameChar(codePoint); next(in))
			name.appendCodePoint(codePoint);
		return name;
	}

	private int next(Reader in) throws IOException {
		codePoint = in.read();
		return Character.isWhitespace(codePoint) ? next(in) : codePoint;
	}

	/**
	 * Override this if you want to use other/more characters. E.g.:<br>
	 * return super.isNameChar(codePoint) || codePoint == '+' || codePoint == '-';
	 */
	public boolean isNameChar(int codePoint) {
		return codePoint == nf.PREFIX_FreshTermVar || codePoint == nf.PREFIX_FreshHedgeVar || codePoint == '_'
				|| Character.isLetterOrDigit(codePoint);
	}
}
