/*
 * Copyright 2012-2018 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.tgau.data;

import java.util.Deque;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import at.jku.risc.stout.tgau.data.atom.FunctionApplication;
import at.jku.risc.stout.tgau.data.atom.FunctionSymbol;
import at.jku.risc.stout.tgau.data.atom.HedgeVar;
import at.jku.risc.stout.tgau.data.atom.Term;
import at.jku.risc.stout.tgau.data.atom.TermAtom;
import at.jku.risc.stout.tgau.data.atom.TermVar;
import at.jku.risc.stout.tgau.data.atom.Variable;
import at.jku.risc.stout.tgau.util.DataStructureFactory;

/**
 * A factory to create {@linkplain Term}s and atomic types (
 * {@linkplain TermAtom}s).
 * 
 * @author Alexander Baumgartner
 */
public class NodeFactory {
	public char PREFIX_FreshTermVar = '&';
	public char PREFIX_FreshHedgeVar = '#';

	public long varCnt = 0;
	private Deque<List<Variable>> hStack = DataStructureFactory.$.newDeque();

	private Map<String, Variable> varCache = new HashMap<String, Variable>();
	private Map<String, FunctionSymbol> fncCache = new HashMap<String, FunctionSymbol>();

	public void pushHedge() {
		hStack.push(DataStructureFactory.$.<Variable>newList());
	}

	public void addToHedge(Variable node) {
		hStack.peek().add(node);
	}

	public List<Variable> popHedge() {
		return hStack.pop();
	}

	public Variable createHedgeVar(String name) throws MalformedTermException {
		Variable var = varCache.get(name);
		if (var == null)
			return cacheIt(new HedgeVar(name));
		return var;
	}

	public Variable createTermVar(String name) throws MalformedTermException {
		Variable var = varCache.get(name);
		if (var == null)
			return cacheIt(new TermVar(name));
		return var;
	}

	private Variable cacheIt(Variable var) throws MalformedTermException {
		String name = var.getName();
		if (name.charAt(0) == PREFIX_FreshTermVar || name.charAt(0) == PREFIX_FreshHedgeVar) {
			long num = parseNum(name.substring(1));
			if (num > varCnt)
				varCnt = num;
		}
		varCache.put(name, var);
		return var;
	}

	private long parseNum(String s) throws MalformedTermException {
		long ret = 0;
		for (int i = 0; i < s.length(); i++) {
			ret *= 10;
			int digit = Character.digit(s.charAt(i), 10);
			if (digit < 0)
				return digit;
			ret += digit;
			if (ret > Long.MAX_VALUE / 900)
				throw new MalformedTermException(
						"Out of fresh variables! Rename the variables in your graph or change the prefix for fresh variables (see NodeFactory).");
		}
		return ret;
	}

	public void setCommutative(String... symbolNames) {
		for (String name : symbolNames)
			getFunctionSymbol(name).setCommutative(true);
	}

	public FunctionApplication createFunction(String name, List<Variable> hedge) {
		return new FunctionApplication(getFunctionSymbol(name), hedge);
	}

	public FunctionApplication createConstant(String name) {
		return createFunction(name, null);
	}

	public FunctionSymbol getFunctionSymbol(String name) {
		FunctionSymbol fnc = fncCache.get(name);
		if (fnc == null) {
			fnc = new FunctionSymbol(name);
			fncCache.put(name, fnc);
		}
		return fnc;
	}

	/**
	 * Fresh variables are not cached. Call {@linkplain #createHedgeVar(String)} if
	 * you need to put the new variable into the cache.
	 * 
	 * @return A fresh {@linkplain HedgeVar hedge variable}
	 */
	public HedgeVar obtainFreshHedgeVar() {
		StringBuilder name = new StringBuilder();
		name.append(PREFIX_FreshHedgeVar);
		name.append(++varCnt);
		return new HedgeVar(name.toString());
	}

	/**
	 * Fresh variables are not cached. Call {@linkplain #createHedgeVar(String)} if
	 * you need to put the new variable into the cache.
	 * 
	 * @return A fresh {@linkplain HedgeVar hedge variable}
	 */
	public TermVar obtainFreshTermVar() {
		StringBuilder name = new StringBuilder();
		name.append(PREFIX_FreshTermVar);
		name.append(++varCnt);
		return new TermVar(name.toString());
	}
}
