/*
 * Copyright 2012-2018 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.tgau.data;

import java.io.IOException;
import java.io.Writer;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Map.Entry;

import at.jku.risc.stout.tgau.data.atom.FunctionApplication;
import at.jku.risc.stout.tgau.data.atom.HedgeVar;
import at.jku.risc.stout.tgau.data.atom.Term;
import at.jku.risc.stout.tgau.data.atom.TermAtom;
import at.jku.risc.stout.tgau.data.atom.TermVar;
import at.jku.risc.stout.tgau.data.atom.Variable;
import at.jku.risc.stout.tgau.util.DataStructureFactory;
import at.jku.risc.stout.tgau.util.PrintableX;

/**
 * A term graph is a system of recursion equations with a predefined root.
 * 
 * @author Alexander Baumgartner
 */
public class TermGraph extends PrintableX implements Comparable<TermGraph> {
	private TermVar root;
	private Map<Variable, Term> recursionEquations;
	private int suggestSize = 0;
	private Variable[] suggestFrom = new Variable[8];
	private Term[] suggestTo = new Term[8];

	public TermGraph(TermVar root) {
		this.root = root;
		this.recursionEquations = DataStructureFactory.$.newMap();
	}

	public TermGraph(TermVar root, Map<Variable, Term> recursionEquations) {
		this.root = root;
		this.recursionEquations = recursionEquations;
	}

	/**
	 * Use and manipulate it carefully! Adding equations that are not in canonical
	 * form leads to unpredictable behavior.
	 * 
	 * @return Map of recursion equations
	 */
	public Map<Variable, Term> getRecursionEquations() {
		return recursionEquations;
	}

	/**
	 * Adds a recursion equation in canonical form to the system of equations of the
	 * term graph.
	 * 
	 * @return see {@linkplain Map#put(Object, Object)}
	 * @throws MalformedTermException
	 *             If the equation is not in canonical form.
	 */
	public Term addRecursionEquation(Variable var, Term node) throws MalformedTermException {
		if (var instanceof HedgeVar != node instanceof HedgeVar)
			throw new MalformedTermException("Equation " + var + " is not in canonical form");
		return recursionEquations.put(var, node);
	}

	/**
	 * Check and remove useless (unreachable) equations
	 * 
	 * @throws MalformedTermException
	 */
	public void checkAndTrim() throws MalformedTermException {
		Map<Variable, Term> rNew = DataStructureFactory.$.newMap(recursionEquations.size());
		if (recursionEquations.get(root) != null)
			addReachable(rNew, root);
		recursionEquations = rNew;
	}

	@Override
	public int hashCode() {
		return (997 * root.hashCode()) ^ (991 * recursionEquations.hashCode());
	}

	@Override
	public boolean equals(Object o) {
		if (!(o instanceof TermGraph))
			return false;
		TermGraph tg2 = (TermGraph) o;
		return root == tg2.root && recursionEquations.equals(tg2.recursionEquations);
	}

	private TermAtom getTopAtom(TermAtom atom) {
		Term t = recursionEquations.get(atom);
		return t == null ? atom : t.getAtom();
	}

	public TermAtomList top() {
		return top(root);
	}

	public TermAtomList top(TermAtom var) {
		TermAtomList list = TermAtomList.obtainList();
		list.add(getTopAtom(var));
		return list;
	}

	public TermAtomList top(List<Variable> recVars) {
		TermAtomList list = TermAtomList.obtainList();
		for (Variable x : recVars)
			list.add(getTopAtom(x));
		return list;
	}

	/**
	 * Sort the outgoing edges of commutative function symbols. This function should
	 * be called once after creating the term graph.
	 */
	public void sortCommutative() {
		for (Entry<Variable, Term> e : recursionEquations.entrySet()) {
			final Term t = e.getValue();
			if (t instanceof FunctionApplication) {
				FunctionApplication appl = (FunctionApplication) t;
				if (appl.getSymbol().isCommutative()) {
					Collections.sort(appl.getArgs(), new Comparator<Variable>() {
						@Override
						public int compare(Variable o1, Variable o2) {
							if (o1 == o2)
								return 0;
							Term t1 = recursionEquations.get(o1);
							Term t2 = recursionEquations.get(o2);
							if (t1 != null && t2 != null) {
								int cmp = t1.getAtom().compare(t2.getAtom());
								if (cmp != 0)
									return cmp;
								if (t1 instanceof FunctionApplication && t2 instanceof FunctionApplication) {
									cmp = ((FunctionApplication) t2).getArgs().size()
											- ((FunctionApplication) t1).getArgs().size();
									if (cmp != 0)
										return cmp;
								}
							}
							return o1.compareTo(o2);
						}
					});
				}
			}
		}
	}

	public TermVar getRoot() {
		return root;
	}

	/**
	 * Just sets the root without removing any equations.
	 * 
	 * @param root
	 */
	public void setRoot(TermVar root) {
		this.root = root;
	}

	/**
	 * Creates a copy with the specified root and removes all useless (unreachable)
	 * equations
	 * 
	 * @param newRoot
	 * @throws MalformedTermException
	 */
	public TermGraph copyRootAndTrim(TermVar newRoot) throws MalformedTermException {
		TermGraph g = new TermGraph(newRoot);
		if (recursionEquations.get(newRoot) != null)
			addReachable(g.recursionEquations, newRoot);
		return g;
	}

	private void addReachable(Map<Variable, Term> recEqNew, Variable x) throws MalformedTermException {
		if (recEqNew.containsKey(x))
			return;
		Term n = recursionEquations.get(x);
		if (n == null || recursionEquations.containsKey(n))
			throw new MalformedTermException("Term graph is not in canonical form");
		recEqNew.put(x, n);
		if (n instanceof FunctionApplication) {
			for (Variable child : ((FunctionApplication) n).getArgs())
				addReachable(recEqNew, child);
		}
	}

	public Term get(Variable recursionVar) {
		return recursionEquations.get(recursionVar);
	}

	public boolean isHorizontallyBound() {
		return true;
	}

	/**
	 * The root is the first printed variable.
	 */
	@Override
	public void print(Writer out) throws IOException {
		out.append('{');
		Term rootN = recursionEquations.get(root);
		if (rootN != null) {
			out.append(root.getName()).append('=');
			rootN.print(out);
			for (Entry<Variable, Term> e : recursionEquations.entrySet()) {
				if (e.getKey() == root)
					continue;
				out.append("; ").append(e.getKey().getName()).append('=');
				e.getValue().print(out);
			}
		}
		out.append('}');
	}

	/**
	 * Replaces one variable by another one.
	 */
	public void substitute(Variable from, Variable to) {
		if (root == from)
			root = (TermVar) to;
		for (Entry<Variable, Term> e : recursionEquations.entrySet()) {
			Term t = e.getValue();
			if (t instanceof FunctionApplication) {
				ListIterator<Variable> i = ((FunctionApplication) t).getArgs().listIterator();
				while (i.hasNext()) {
					if (i.next() == from)
						i.set(to);
				}
			} else if (t == from) {
				e.setValue(to);
			}
		}
	}

	/**
	 * Replaces from by to.
	 * 
	 * @param from
	 *            A hedge variable.
	 * @param to
	 *            A list of variables.
	 */
	public void substitute(Variable from, List<Variable> to) {
		if (to.size() == 1) {
			substitute(from, to.get(0)); // faster
			return;
		}
		for (Entry<Variable, Term> e : recursionEquations.entrySet()) {
			if (e.getValue() instanceof FunctionApplication) {
				List<Variable> args = ((FunctionApplication) e.getValue()).getArgs();
				for (int i = args.size() - 1; i >= 0; i--) {
					if (args.get(i) == from) {
						args.remove(i);
						args.addAll(i, to);
					}
				}
			}
		}
	}

	/**
	 * Suggests a mapping from a variable to a term node. The suggested mapping is
	 * NOT in the term graph until {@link #commitAllMappings()} is executed.
	 * Suggested mappings can be canceled by calling {@link #resetAllMappings()}.
	 */
	public void suggestMapping(Variable from, Term to) {
		if (suggestSize == suggestFrom.length) {
			suggestFrom = Arrays.copyOf(suggestFrom, suggestSize + 8);
			suggestTo = Arrays.copyOf(suggestTo, suggestSize + 8);
		}
		suggestFrom[suggestSize] = from;
		suggestTo[suggestSize] = to;
		suggestSize++;
	}

	public void commitAllMappings() {
		while (suggestSize > 0) {
			suggestSize--;
			recursionEquations.put(suggestFrom[suggestSize], suggestTo[suggestSize]);
		}
	}

	public void resetAllMappings() {
		suggestSize = 0;
	}

	public TermGraph copy() {
		TermGraph gCopy = new TermGraph(root, DataStructureFactory.$.newMap(recursionEquations));
		for (Map.Entry<Variable, Term> e : gCopy.recursionEquations.entrySet()) {
			if (e.getValue() instanceof FunctionApplication) {
				FunctionApplication f = (FunctionApplication) e.getValue();
				e.setValue(new FunctionApplication(f.getSymbol(), DataStructureFactory.$.newList(f.getArgs())));
			}
		}
		return gCopy;
	}

	public void merge(Variable keep, Variable remove) {
		Iterator<Entry<Variable, Term>> i = recursionEquations.entrySet().iterator();
		Variable keepKey = null, removeKey = null;
		while (i.hasNext()) {
			Entry<Variable, Term> e = i.next();
			if (e.getValue() == keep) {
				keepKey = e.getKey();
				if (removeKey != null)
					break;
			} else if (e.getValue() == remove) {
				removeKey = e.getKey();
				i.remove();
				if (keepKey != null)
					break;
			}
		}
		substitute(removeKey, keepKey);
	}

	/**
	 * Removes one mapping for a given value and returns the key. If there is no
	 * mapping to the given value then null is returned.
	 */
	public Variable removeByValue(Variable termValue) {
		Iterator<Entry<Variable, Term>> i = recursionEquations.entrySet().iterator();
		while (i.hasNext()) {
			Entry<Variable, Term> e = i.next();
			if (e.getValue() == termValue) {
				i.remove();
				return e.getKey();
			}
		}
		return null;
	}

	/**
	 * Poor performance - do not use if not absolutely necessary.
	 */
	@Override
	public int compareTo(TermGraph o) {
		int comp = root.compare(o.root);
		if (comp != 0)
			return comp;
		comp = recursionEquations.hashCode() - o.recursionEquations.hashCode();
		if (comp != 0)
			return comp;
		if (recursionEquations.equals(o.recursionEquations))
			return 0;
		return toString().compareTo(o.toString());
	}

}
