/*
 * Copyright 2012-2018 Alexander Baumgartner
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package at.jku.risc.stout.tgau.data.atom;

import java.io.IOException;
import java.io.Writer;
import java.util.Collections;
import java.util.List;

import at.jku.risc.stout.tgau.util.DataStructureFactory;
import at.jku.risc.stout.tgau.util.PrintableX;

/**
 * A function application consists of a {@linkplain FunctionSymbol} and a
 * {@linkplain List} of variables.
 * 
 * @author Alexander Baumgartner
 */
public class FunctionApplication extends PrintableX implements Term {
	private FunctionSymbol symbol;
	private List<Variable> args;

	/**
	 * Instantiates a new term node. If args is null it will be set to the empty
	 * list. So that no {@linkplain NullPointerException} can occur.
	 */
	public FunctionApplication(FunctionSymbol symbol, List<Variable> args) {
		setSymbol(symbol);
		setArgs(args);
	}

	public FunctionSymbol getAtom() {
		return symbol;
	}

	public FunctionSymbol getSymbol() {
		return symbol;
	}

	public void setSymbol(FunctionSymbol atom) {
		this.symbol = atom;
	}

	public List<Variable> getArgs() {
		return args;
	}

	/**
	 * Null values are transformed to unique objects which represent the null value.
	 */
	public void setArgs(List<Variable> args) {
		if (args == null)
			this.args = Collections.emptyList();
		else
			this.args = args;
	}

	@Override
	public int compareTo(Term o) {
		int comp = symbol.compare(o.getAtom());
		if (comp != 0)
			return comp;
		FunctionApplication oAppl = (FunctionApplication) o;
		comp = args.size() - oAppl.args.size();
		for (int i = args.size() - 1; comp != 0 && i >= 0; i--)
			comp = args.get(i).compare(oAppl.args.get(i));
		return comp;

	}

	@Override
	public int hashCode() {
		return (83 * symbol.hashCode()) ^ (89 * args.hashCode());
	}

	@Override
	public boolean equals(Object other) {
		if (!(other instanceof FunctionApplication))
			return false;
		FunctionApplication oAppl = (FunctionApplication) other;
		if (symbol != oAppl.symbol)
			return false;
		if (args == oAppl.args)
			return true;
		return args.equals(oAppl.args);
	}

	@Override
	public void print(Writer out) throws IOException {
		symbol.print(out);
		if (args.size() > 0)
			DataStructureFactory.$.printCollection(args, symbol.isCommutative(), out);
	}
}
